/*	$NetBSD: epclk.c,v 1.3 2005/01/06 16:11:54 joff Exp $	*/

/*
 * Copyright (c) 2004 Jesse Off
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the NetBSD
 *	Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Driver for the ep93xx clock tick.
 *
 * We use the 64Hz RTC interrupt as its the only thing that allows for timekeeping
 * of a second (crystal error only).  There are two general purpose timers
 * on the ep93xx, but they run at a frequency that makes a perfect integer 
 * number of ticks per second impossible.  Note that there was an errata with
 * the ep93xx processor and many early boards (including the Cirrus eval board) have
 * a broken crystal oscillator input that may make this 64Hz unreliable.  However,
 * not all boards are susceptible, the Technologic Systems TS-7200 is a notable
 * exception that is immune to this errata.   --joff
 */

/*
 * $NetBSD: epclk.c,v 1.3 2005/01/06 16:11:54 joff Exp $
 */
#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/time.h>
#include <sys/module.h>
#include <sys/rman.h>
#include <sys/bus.h>
#include <machine/bus.h>
#include <machine/intr.h>

#include <machine/cpufunc.h>

#include <arm/ep93xx/epclkreg.h> 
#include <arm/ep93xx/ep93xxvar.h>

static int	epclk_probe(device_t);
static void	epclk_attach(device_t);

/* callback functions for intr_functions */
static void      epclk_intr(void* arg);


/* This is a quick ARM way to multiply by 983040/1000000 */
#define US_TO_TIMER4VAL(x) { \
	u_int32_t hi, lo, scalar = 4222124650UL; \
	__asm __volatile ( \
		"umull %0, %1, %2, %3;" \
		: "=&r"(lo), "=&r"(hi) \
		: "r"((x)), "r"(scalar) \
	); \
	(x) = hi; \
}

/* This is a quick ARM way to multiply by 1000000/983040 */
#define TIMER4VAL_TO_US(x) { \
	u_int32_t hi, lo, scalar = 2184533333UL; \
	asm volatile ( \
		"umull %0, %1, %2, %3;" \
		"mov %1, %1, lsl #1;" \
		"mov %0, %0, lsr #31;" \
		"orr %1, %1, %0;" \
		: "=&r"(lo), "=&r"(hi) \
		: "r"((x)), "r"(scalar) \
	); \
	(x) = hi; \
}


#define TIMER4VAL()	(*(volatile u_int32_t *)(EP93XX_APB_VBASE + \
	EP93XX_APB_TIMERS + EP93XX_TIMERS_Timer4ValueLow))

static int
epclk_probe(device_t dev)
{

	return 0;
}

struct epclk_softc {
	bus_space_tag_t sc_iot;
	bus_space_handle_t sc_ioh;
	bus_space_handle_t sc_teoi_ioh;
	device_t dev;
};

static struct epclk_softc *epclk;

static void
epclk_attach(device_t dev)
{

	epclk = device_get_softc(dev);
	epclk->sc_iot = &ep93xx_bs_tag;
	epclk->sc_ioh = EP93XX_APB_VBASE;
	epclk->dev = dev;
	epclk->sc_teoi_ioh = EP93XX_APB_VBASE + EP93XX_APB_SYSCON +
	    EP93XX_SYSCON_TEOI;
	/* clear and start the debug timer (Timer4) */
	bus_space_write_4(epclk->sc_iot, epclk->sc_ioh,
	    EP93XX_TIMERS_Timer4Enable, 0);
	bus_space_write_4(epclk->sc_iot, epclk->sc_ioh,
	    EP93XX_TIMERS_Timer4Enable, 0x100);
}

static device_method_t epclk_methods[] = {
	DEVMETHOD(device_probe, epclk_probe),
	DEVMETHOD(device_attach, epclk_attach),
	{0, 0},
};

static driver_t epclk_driver = {
	"epclk",
	epclk_methods,
	sizeof(struct epclk_softc),
};
static devclass_t epclk_devclass;

DRIVER_MODULE(epclk, eparm, epclk_driver, epclk_devclass, 0, 0);
/*
 * epclk_intr:
 *
 *	Handle the hardclock interrupt.
 */
static void
epclk_intr(void *arg)
{

	bus_space_write_4(epclk->sc_iot, epclk->sc_teoi_ioh, 0, 1);
	hardclock((struct clockframe*) arg);
}


/*
 * cpu_initclocks:
 *
 *	Initialize the clock and get them going.
 */
void
cpu_initclocks(void)
{
	struct resource *irq;
	void *ihl;
	int rid = 0;

	stathz = profhz = 0;

	if (hz != 64) panic("HZ must be 64!");

	/* clear 64Hz interrupt status */
	bus_space_write_4(epclk->sc_iot, epclk->sc_teoi_ioh, 0, 1);

	irq = bus_alloc_resource(epclk->dev, SYS_RES_IRQ, &rid, 35,
	    35, 1, RF_ACTIVE);
	if (!irq)
		panic("Unable to setup the clock irq handler.\n");
	else
		bus_setup_intr(epclk->dev, irq, INTR_TYPE_CLK | INTR_FAST, 
		    epclk_intr, NULL, &ihl);
}

/*
 * DELAY:
 *
 *	Delay for at least N microseconds.
 */
void
DELAY(int len)
{
	u_int32_t	start, end, ticks;

#ifdef DEBUG
	if (epclk_sc == NULL) {
		printf("delay: called before start epclk\n");
		return;
	}
#endif

	ticks = start = TIMER4VAL();
	US_TO_TIMER4VAL(len);
	end = start + len;
	while (start <= ticks && ticks > end) {
		/* wait for Timer4ValueLow wraparound */
		ticks = TIMER4VAL();
	}
	while (ticks <= end) {
		ticks = TIMER4VAL();
	}
}

void
cpu_startprofclock()
{
}

void
cpu_stopprofclock()
{
}
